import React, { useState, useContext,createContext } from 'react';



export interface AuthBuilder {
  getToke:() => string | null;
  setToken:(token:string) => void;
  removeToken:() => void;
}




interface Token {
  token:string;
}



export interface AuthStates<TUser extends {token:string}> {
  isLogin:boolean;
  user:TUser | null;
}


export interface AuthContextType<TUser extends {token:string}> extends AuthStates<TUser> {
  login:(user:TUser) => void;
  logout:() => void;
}





const defaultAuthStoreArg:AuthBuilder = {
  getToke:() => null,
  setToken:(token) => {},
  removeToken:() => {}
}




export const authStoreBuilder = function<TUser extends Token>(authBuilderArg:AuthBuilder = defaultAuthStoreArg){

  const {getToke,setToken,removeToken} = authBuilderArg;

  const AuthContext = createContext<AuthContextType<TUser>>({
    isLogin:false,
    user:null,
    login(user) {
        
    },
    logout() {
        
    },
  });
  

  const token = getToke();

  
  const AuthProvider = (props:React.PropsWithChildren) => {
    
    const [auth,setAuth] = useState<AuthStates<TUser>>({
      isLogin:!!token,
      user:null
    });
    
    function login(user:TUser){
      setToken(user.token);
  
      setAuth({isLogin:true,user:user});
    }


    function logout(){
      removeToken();
      setAuth({isLogin:false,user:null});
    }


    return (
      <AuthContext.Provider value={{
        ...auth,
        login,
        logout
      }}>
        {props.children}
      </AuthContext.Provider>
    )
  }



  function useAuthStore(){
    return useContext(AuthContext);
  }


  return {
    AuthProvider,
    useAuthStore,
  }

}





